use std::sync::Arc;

use axum::{
    http,
    response::{Html, IntoResponse},
    routing::get,
    Router,
};
use axum_typed_websockets::{Message, WebSocket, WebSocketUpgrade};
use serde::{Deserialize, Serialize};

use offline_web_model::{Graph, Reference};

async fn get_client_js() -> impl IntoResponse {
    (
        [(http::header::CONTENT_TYPE, "application/javascript")],
        include_str!("../static/client.js"),
    )
}

#[derive(Debug, Serialize, Deserialize)]
enum ServerMsg {
    Reference(Reference),
    Object(Vec<u8>),
}

#[derive(Debug, Serialize, Deserialize)]
enum ClientMsg {
    Bootstrap,
    GetReference(String),
    GetObject(String),
}

async fn handle_websocket(
    ws: WebSocketUpgrade<ServerMsg, ClientMsg>,
    graph: Arc<Graph>,
) -> impl IntoResponse {
    ws.on_upgrade(|socket| handle_socket(socket, graph))
}

async fn handle_socket(mut socket: WebSocket<ServerMsg, ClientMsg>, graph: Arc<Graph>) {
    // Send initial data to the client

    // Handle incoming messages
    // TODO(zaphar): This should actually be smarter in the case of None or Err.
    while let Some(Ok(_msg)) = socket.recv().await {
        match _msg {
            Message::Item(ClientMsg::Bootstrap) => {
                if let Err(e) = socket
                    .send(Message::Item(ServerMsg::Reference(
                        graph.root.as_ref().clone(),
                    )))
                    .await
                {
                    println!("Error sending bootstrap reference: {:?}", e);
                    continue;
                }
            }
            Message::Item(ClientMsg::GetReference(path)) => {
                if let Some(reference) = graph.refs.get(&path) {
                    if let Err(e) = socket
                        .send(Message::Item(ServerMsg::Reference((**reference).clone())))
                        .await
                    {
                        println!("Error sending reference: {:?}", e);
                        continue;
                    }
                } else {
                    println!("Reference not found: {}", path);
                }
            }
            Message::Item(ClientMsg::GetObject(address)) => {
                if let Some(object) = graph.get_object(&address) {
                    if let Err(e) = socket
                        .send(Message::Item(ServerMsg::Object(object.clone())))
                        .await
                    {
                        println!("Error sending object: {:?}", e);
                        continue;
                    }
                } else {
                    println!("Object not found: {}", address);
                }
            }
            Message::Ping(items) => {
                println!("unhandled ping msg: {:?}", items);
            }
            Message::Pong(items) => {
                println!("unhandled pong msg: {:?}", items);
            }
            Message::Close(_close_frame) => {
                println!("closing websocket connection at client request");
                break;
            }
        }
    }

    println!("WebSocket connection closed");
}

pub fn endpoints(graph: Graph) -> Router {
    // WebSocket route now implemented
    let graph = Arc::new(graph);
    Router::new()
        .nest(
            "/api/v1",
            Router::new().route(
                "/ws",
                get(|ws: WebSocketUpgrade<ServerMsg, ClientMsg>| async move {
                    handle_websocket(ws, graph.clone()).await
                }),
            ),
        )
        .route("/lib/client.js", get(get_client_js))
        .route(
            "/ui/",
            get(|| async { Html(include_str!("../static/index.html")).into_response() }),
        )
}

// TODO(jwall): Javascript test script
pub async fn serve() {
    // run our app with hyper, listening globally on port 3000
    let graph = Graph::random_graph();
    let listener = tokio::net::TcpListener::bind("127.0.0.1:3000")
        .await
        .unwrap();
    println!("Server ui starting on http://127.0.0.1:3000/ui/");
    println!("WebSocket endpoint available at ws://127.0.0.1:3000/api/v1/ws");
    axum::serve(listener, endpoints(graph)).await.unwrap();
}
